import gym
import numpy
from numpy import random
from numba import njit
from l3c.anymdp.solver import update_value_matrix


class AnyMDPSolverOpt(object):
    """
    Solver for AnyMDPEnv with Bellman Equation and Value Iteration
    Suppose to know the ground truth of the environment
    """
    def __init__(self, env, gamma=0.98):
        if(not env.task_set):
            raise Exception("AnyMDPEnv is not initialized by 'set_task', must call set_task first")
        self.n_actions = env.action_space.n
        self.n_states = env.observation_space.n
        self.transition_matrix = env.transition_matrix
        self.reward_matrix = env.reward_matrix
        self.value_matrix = numpy.zeros((self.n_states, self.n_actions))
        self.gamma = gamma
        self.q_solver()

    def q_solver(self, gamma=0.99):
        self.value_matrix = update_value_matrix(self.reward_matrix, self.transition_matrix, gamma, self.value_matrix)

    def policy(self, state):
        return numpy.argmax(self.value_matrix[state])


import gym
import numpy
from numpy import random
from numba import njit
import networkx as nx


@njit(cache=True)
def update_value_matrix(r_mat, t_mat, gamma, vm, max_iteration=-1):
    diff = 1.0
    cur_vm = numpy.copy(vm)
    ns, na = r_mat.shape
    iteration = 0

    while diff > 1.0e-4 and (
            (max_iteration < 0) or 
            (max_iteration > iteration and max_iteration > 1) or
            (iteration < 1 and random.random() < max_iteration)):
        iteration += 1
        old_vm = numpy.copy(cur_vm)
        for s in range(ns):
            for a in range(na):
                exp_q = 0.0
                for sn in range(ns):
                    exp_q += t_mat[s,a,sn] * numpy.max(cur_vm[sn])
                cur_vm[s,a] = r_mat[s,a] + gamma * exp_q
        diff = numpy.sqrt(numpy.mean((old_vm - cur_vm)**2))
    return cur_vm

def check_transition_quality(transition):
    if(transition is None):
        return False
    # acquire state - to - state distribution
    in_dist = numpy.sum(transition, axis=1)

    in_graph = nx.DiGraph()
    nodes = range(in_dist.shape[0])
    in_graph.add_nodes_from(nodes)
    for i in range(in_dist.shape[0]):
        for j in range(in_dist.shape[1]):
            if in_dist[i, j] != 0:
                in_graph.add_edge(i, j, weight=in_dist[i, j])

    if(nx.is_strongly_connected(in_graph)):
        return True
    else:
        print("AnyMDP: Sampling transition graph not strongly connected, need resample...")
        return False

def check_reward_quality(transition, reward):
    if(transition is None or reward is None):
        return False
    ns, na = reward.shape
    vm = update_value_matrix(reward, transition, 0.98, numpy.zeros((ns, na), dtype=float), max_iteration=2)
    if(numpy.std(vm) > 1.0e-2 and numpy.min(vm) < 0 and numpy.max(vm) > 0):
        return True
    else:
        print("AnyMDP: Sampling task has trivial value function, need resample...")
        return False

def check_task_quality(task):
    """
    Check the quality of the task
    Requiring: Q value is diverse enough
               State is connected enough (no isolated state)
    """

    if(task is None):
        return False
    if(check_transition_quality(task["transition"]) and 
            check_reward_quality(task["transition"], task["reward"])):
        return True
    return False





import numpy
import pygame
import time
import random
from numba import njit
from l3c.mazeworld.envs.dynamics import PI, PI_4

class LandmarksRGB(object):
    def __init__(self):
        self._landmarks_rgb = dict()
        self._landmarks_rgb[0] = numpy.array([0, 255, 0], dtype="float32") # Green
        self._landmarks_rgb[1] = numpy.array([255, 0, 0], dtype="float32") # Red
        self._landmarks_rgb[2] = numpy.array([0, 0, 255], dtype="float32") # Blue
        self._landmarks_rgb[3] = numpy.array([0, 255, 255], dtype="float32") # 
        self._landmarks_rgb[4] = numpy.array([255, 0, 255], dtype="float32") #
        self._landmarks_rgb[5] = numpy.array([255, 255, 0], dtype="float32") #
        self._landmarks_rgb[6] = numpy.array([128, 128, 255], dtype="float32") #
        self._landmarks_rgb[7] = numpy.array([128, 255, 128], dtype="float32") #
        self._landmarks_rgb[8] = numpy.array([255, 128, 128], dtype="float32") #
        self._landmarks_rgb[9] = numpy.array([0, 96, 128], dtype="float32") #
        self._landmarks_rgb[10] = numpy.array([96, 0, 128], dtype="float32") #
        self._landmarks_rgb[11] = numpy.array([0, 128, 96], dtype="float32") #
        self._landmarks_rgb[12] = numpy.array([96, 128, 0], dtype="float32") #
        self._landmarks_rgb[13] = numpy.array([128, 96, 0], dtype="float32") #
        self._landmarks_rgb[14] = numpy.array([128, 0, 96], dtype="float32") #

    def color(self, i, opacity=0.0):
        real_color = (1 - opacity) * self._landmarks_rgb[i] + opacity * numpy.array([255, 255, 255])
        return pygame.Color(int(real_color[0]), int(real_color[1]), int(real_color[2]), 255)

    @property
    def rgb(self):
        return self._landmarks_rgb

    @property
    def rgb_npa(self):
        rgb_arr = numpy.zeros((len(self._landmarks_rgb), 3), dtype="float32")
        for i in self._landmarks_rgb:
            rgb_arr[i] = self._landmarks_rgb[i]
        return rgb_arr

lrgb_handler = LandmarksRGB()
landmarks_rgb = lrgb_handler.rgb
landmarks_color = lrgb_handler.color
landmarks_rgb_arr = lrgb_handler.rgb_npa

@njit(cache=True)
def DDA_2D(pos, i, j, cell_number, cell_size, cos_ori, sin_ori, cell_walls, cell_transparent, visibility_3D):
    eps = 1.0e-8
    if(cos_ori < 0):
        c_sign = -1
    else:
        c_sign = 1
    if(sin_ori < 0):
        s_sign = -1
    else:
        s_sign = 1
    delta_dist_x = abs(cell_size / eps) if abs(cos_ori) < eps else abs(cell_size / cos_ori)
    delta_dist_y = abs(cell_size / eps) if abs(sin_ori) < eps else abs(cell_size / sin_ori)
    d_x = ((i + 1) * cell_size - pos[0]) if cos_ori > 0 else (i * cell_size - pos[0])
    d_y = ((j + 1) * cell_size - pos[1]) if sin_ori > 0 else (j * cell_size - pos[1])
    side_dist_x = c_sign * (d_x / eps) if abs(cos_ori) < eps else d_x / cos_ori
    side_dist_y = s_sign * (d_y / eps) if abs(sin_ori) < eps else d_y / sin_ori
    delta_i = 1 if(cos_ori > 0) else -1
    delta_j = 1 if(sin_ori > 0) else -1
    hit_i = i
    hit_j = j
    hit_dist = 0.0
    hit_side = 0
    hit_transparent_list = []
    exposed_cell = [[i,j]]

    #Remove this part as we don't want to see colors when in landmark cells themselves
    #if(cell_transparent[hit_i, hit_j] > -1):
    #    if(side_dist_x < side_dist_y):
    #        hit_transparent_list.append((side_dist_x, hit_i, hit_j, 0, cell_transparent[hit_i, hit_j]))
    #    elif(side_dist_x > side_dist_y):
    #        hit_transparent_list.append((side_dist_y, hit_i, hit_j, 1, cell_transparent[hit_i, hit_j]))

    while hit_dist < visibility_3D:
        if(side_dist_x < side_dist_y):
            hit_i += delta_i
            side_dist_y -= side_dist_x
            hit_dist += side_dist_x
            if(cell_transparent[hit_i, hit_j] > -1):
                hit_transparent_list.append((hit_dist, hit_i, hit_j, 0, cell_transparent[hit_i, hit_j]))
            if(hit_i < 0 or hit_i >= cell_number):
                if(hit_j < 0 or hit_j >= cell_number):
                    hit_dist = 1.0e+6
                    break
            else:
                if(hit_dist <= visibility_3D * 0.60):
                    exposed_cell.append([hit_i, hit_j]) # The cell becomes seen to the agent
                if(cell_walls[hit_i, hit_j] > 0):
                    hit_side = 0
                    break
            side_dist_x = delta_dist_x
        else:
            hit_j += delta_j
            side_dist_x -= side_dist_y
            hit_dist += side_dist_y
            if(cell_transparent[hit_i, hit_j] > -1):
                hit_transparent_list.append((hit_dist, hit_i, hit_j, 1, cell_transparent[hit_i, hit_j]))
            if(hit_i < 0 or hit_i >= cell_number):
                if(hit_j < 0 or hit_j >= cell_number):
                    hit_dist = 1.0e+6
                    break
            else:
                if(hit_dist <= visibility_3D * 0.60):
                    exposed_cell.append([hit_i, hit_j]) # The cell becomes seen to the agent
                if(cell_walls[hit_i, hit_j] > 0):
                    hit_side = 1
                    break
            side_dist_y = delta_dist_y
    return hit_dist, hit_i, hit_j, hit_side, hit_transparent_list, exposed_cell


"""
Depict 3D Maze View
cell transparent: N X N array, where -1 represents empty, and 0~9 (max = 9 represents different colors of transparency)
"""

@njit(cache=True)
def interpolate(array, i, j, d, px, py):
    d2 = max(d ** 2, 1.0e-8)
    w,h,_ = array.shape
    ib = int(i)
    jb = int(j)
    sum_wht = 0.0
    sum_r = numpy.zeros(shape=(3), dtype="float32")

    for x in range(ib-1, ib+3):
        for y in range(jb-1, jb+3):
            dist = (((x - i) * px) ** 2 + ((y - j) * py) ** 2)
            wht = max(0.01, min(1.0 - 10 * dist / d2, 1.0))
            sum_wht += wht
            xv = x % w
            yv = y % h
            sum_r += wht * array[xv, yv]
    return sum_r / sum_wht

@njit(cache=True)
def maze_view(pos, ori, vision_height, cell_walls, cell_transparent, cell_texts, cell_size, texture_array, ground_text,
        ceil_text, ceil_height, text_size, visibility_3D, l_focal, vision_angle_h, resolution_h, resolution_v, transparent_rgb):
    vision_screen_half_size_h = numpy.tan(vision_angle_h / 2) * l_focal
    vision_screen_half_size_v = vision_screen_half_size_h * resolution_v / resolution_h
    pixel_size = 2.0 * vision_screen_half_size_h / resolution_h
    s_ori = numpy.sin(ori)
    c_ori = numpy.cos(ori)
    max_cell_i = cell_walls.shape[0]
    max_cell_j = cell_walls.shape[1]
    pixel_factor = pixel_size / l_focal
    cell_exposed = numpy.zeros_like(cell_walls)
    percell_textnum = cell_size / text_size
    height_texts = ceil_height / text_size

    FAR_RGB = numpy.array([1, 1, 1], dtype="float32")

    textg_pixel_size_x = text_size / ground_text.shape[0]
    textg_pixel_size_y = text_size / ground_text.shape[1]
    textc_pixel_size_x = text_size / ceil_text.shape[0]
    textc_pixel_size_y = text_size / ceil_text.shape[1]

    # prepare some maths
    rgb_array = numpy.zeros(shape=(resolution_h, resolution_v, 3), dtype="int32")
    rgb_array[:, :] = FAR_RGB
    cos_hp_array = numpy.zeros(shape=(resolution_h,), dtype="float32")
    cos_abs_hp_array = numpy.zeros(shape=(resolution_h,), dtype="float32")
    sin_abs_hp_array = numpy.zeros(shape=(resolution_h,), dtype="float32")
    tan_hp = (- 0.5 - resolution_h / 2) * pixel_factor
    for d_h in range(resolution_h):
        tan_hp += pixel_factor
        cos_hp = numpy.sqrt(1.0 / (1.0 + tan_hp ** 2))
        sin_hp = tan_hp * cos_hp
        sin_abs_hp_array[d_h] = sin_hp * c_ori + cos_hp * s_ori
        cos_abs_hp_array[d_h] = cos_hp * c_ori - sin_hp * s_ori
        cos_hp_array[d_h] = cos_hp
    
    # paint floor
    for d_v in range(resolution_v - 1, resolution_v//2, -1):
        # horizontal distance on screen
        v_screen = (d_v + 0.5) * pixel_size - vision_screen_half_size_v
        distance = vision_height / v_screen * l_focal
        light_incident = min(1.0, v_screen / l_focal)
        if(distance > visibility_3D):
            continue

        for d_h in range(resolution_h):
            eff_distance = distance / cos_hp_array[d_h]
            alpha = min(1.0, max(2.0 * eff_distance / visibility_3D - 1.0, 0.0)) * light_incident
            hit_x = eff_distance * cos_abs_hp_array[d_h] + pos[0]
            hit_y = eff_distance * sin_abs_hp_array[d_h] + pos[1]
            i = (hit_x / cell_size)
            j = (hit_y / cell_size)
            d_i = i - numpy.floor(i)
            d_j = j - numpy.floor(j)
            i = int(i)
            j = int(j)
            eff_ps = eff_distance * pixel_size / l_focal

            if(i < max_cell_i and i >= 0 and j < max_cell_j and j >= 0):
                #text_id = cell_texts[i,j]
                d_i *= percell_textnum
                d_j *= percell_textnum
                d_i -= numpy.floor(d_i)
                d_j -= numpy.floor(d_j)
                d_i *= ground_text.shape[0]
                d_j *= ground_text.shape[1]
                rgb_array[d_h, d_v, :] = numpy.clip(
                        light_incident * (alpha * FAR_RGB + (1.0 - alpha) * interpolate(ground_text, d_i, d_j, eff_ps, textg_pixel_size_x, textg_pixel_size_y)),
                        0, 255)

    # paint ceil
    for d_v in range(resolution_v//2):
        v_screen = vision_screen_half_size_v - (d_v + 0.5) * pixel_size
        distance = (ceil_height - vision_height) / v_screen * l_focal
        light_incident = min(1.0, v_screen / l_focal)
        if(distance > visibility_3D):
            continue

        for d_h in range(resolution_h):
            eff_distance = distance / cos_hp_array[d_h]
            alpha = min(1.0, max(2.0 * eff_distance / visibility_3D - 1.0, 0.0))
            hit_x = eff_distance * cos_abs_hp_array[d_h] + pos[0]
            hit_y = eff_distance * sin_abs_hp_array[d_h] + pos[1]
            i = (hit_x / cell_size)
            j = (hit_y / cell_size)
            d_i = i - numpy.floor(i)
            d_j = j - numpy.floor(j)
            i = int(i)
            j = int(j)
            eff_ps = eff_distance * pixel_size / l_focal

            if(i < max_cell_i and i >= 0 and j < max_cell_j and j >= 0):
                #text_id = cell_texts[i,j]
                d_i *= percell_textnum
                d_j *= percell_textnum
                d_i -= numpy.floor(d_i)
                d_j -= numpy.floor(d_j)
                d_i *= ceil_text.shape[0]
                d_j *= ceil_text.shape[1]
                rgb_array[d_h, d_v, :] = numpy.clip(
                    light_incident * (alpha * FAR_RGB + (1.0 - alpha) * interpolate(ceil_text, d_i, d_j, eff_ps, textc_pixel_size_x, textc_pixel_size_y)),
                    0, 255)
    
    # paint wall
    for d_h in range(resolution_h):
        i = int(pos[0] / cell_size)
        j = int(pos[1] / cell_size)
        hit_dist, hit_i, hit_j, hit_side, hit_transparent, exposed_cell = DDA_2D(
                pos, i, j, max_cell_i, cell_size, cos_abs_hp_array[d_h], sin_abs_hp_array[d_h], 
                cell_walls, cell_transparent, visibility_3D)
        for idx in exposed_cell:
            if(random.random() < 0.05):
                cell_exposed[idx[0],idx[1]] = 1
        hit_transparent.sort(key=lambda x:x[0], reverse=True)

        alpha = min(1.0, max(2.0 * hit_dist / visibility_3D - 1.0, 0.0))
        text_id = cell_texts[hit_i,hit_j]
        hit_pt_x = hit_dist * cos_abs_hp_array[d_h] + pos[0]
        hit_pt_y = hit_dist * sin_abs_hp_array[d_h] + pos[1]
        if(hit_side == 0): # hit vertical wall
           local_h = hit_pt_y / cell_size 
           local_h -= numpy.floor(local_h)
           light_incident = abs(cos_abs_hp_array[d_h])
        else:
           local_h = hit_pt_x / cell_size 
           local_h -= numpy.floor(local_h)
           light_incident = abs(sin_abs_hp_array[d_h])

        ratio = hit_dist * cos_hp_array[d_h] / l_focal
        if(abs(ratio) < 1.0e-8):
            if(ratio > 0):
                ratio = 1.0e-8
            else:
                ratio = -1.0e-8
        top_v = (ceil_height - vision_height) / ratio
        bot_v = vision_height / ratio

        v_s = max(0, int((vision_screen_half_size_v - top_v) / pixel_size))
        v_e = min(resolution_v, int((vision_screen_half_size_v + bot_v) / pixel_size))

        for d_v in range(v_s, v_e):
            local_v = (vision_screen_half_size_v - (d_v + 0.5) * pixel_size) * ratio + vision_height
            d_i = local_h * percell_textnum
            d_j = local_v / text_size
            d_i -= numpy.floor(d_i)
            d_j -= numpy.floor(d_j)
            d_i = int(texture_array[text_id].shape[0] * d_i)
            d_j = int(texture_array[text_id].shape[1] * d_j)
            text_pixel_size_x = text_size / texture_array[text_id].shape[0]
            text_pixel_size_y = text_size / texture_array[text_id].shape[1]

            eff_ps = eff_distance * pixel_size / l_focal

            rgb_array[d_h, d_v, :] = numpy.clip(
                    light_incident * (alpha * FAR_RGB + (1.0 - alpha) * interpolate(texture_array[text_id], d_i, d_j, eff_ps, text_pixel_size_x, text_pixel_size_y)),
                    0, 255)

        # Add those transparent
        for hit_dist, hit_i, hit_j, hit_side, trans_ID in hit_transparent:
            ratio = hit_dist * cos_hp_array[d_h] / l_focal
            if(abs(ratio) < 1.0e-8):
                if(ratio > 0):
                    ratio = 1.0e-8
                else:
                    ratio = -1.0e-8
            top_v = (ceil_height - vision_height) / ratio
            bot_v = vision_height / ratio
            transparent_factor = 0.30

            v_s = max(0, int((vision_screen_half_size_v - top_v) / pixel_size))
            v_e = min(resolution_v, int((vision_screen_half_size_v + bot_v) / pixel_size))
            alpha = min(1.0, max(2.0 * hit_dist / visibility_3D - 1.0, 0.0))
            for d_v in range(v_s, v_e):
                rgb_array[d_h, d_v] = numpy.clip(
                    (1.0 - transparent_factor) * rgb_array[d_h, d_v] + transparent_factor * ((1.0 - alpha) * transparent_rgb[trans_ID] + alpha * FAR_RGB),
                    0, 255)
        
    return rgb_array, cell_exposed

def paint_agent_arrow(scr, color, offset, pos, angle, l1, l2):
    A_ARR = PI / 3
    A_A = angle - (PI - A_ARR)
    A_B = angle + (PI - A_ARR)
    pos_S = (float(pos[0] + offset[0]), float(pos[1] + offset[1]))
    pos_A = (float(pos[0] + l1 * numpy.cos(A_A) + offset[0]), float(pos[1] + l1 * numpy.sin(A_A) + offset[1]))
    pos_B = (float(pos[0] + l1 * numpy.cos(A_B) + offset[0]), float(pos[1] + l1 * numpy.sin(A_B) + offset[1]))
    pos_C = (float(pos[0] + l2 * numpy.cos(angle) + offset[0]), float(pos[1] + l2 * numpy.sin(angle) + offset[1]))
    pygame.draw.polygon(scr, color, (pos_S, pos_A, pos_C, pos_B))



import gym
import sys
import argparse
import time
import l3c.mazeworld
from l3c.mazeworld import MazeTaskSampler
from l3c.mazeworld.agents import SmartSLAMAgent

if __name__=='__main__':
    parser = argparse.ArgumentParser(description='Playing the maze world demo with your keyboard')
    parser.add_argument('--max_steps', type=int, default=1000000)
    parser.add_argument('--visibility_3D', type=float, default=12, help="3D vision range, Only valid in 3D mode")
    parser.add_argument('--save_replay', type=str, default=None, help="Save the replay trajectory in file")
    parser.add_argument('--memory_keep_ratio', type=float, default=1.0, 
                        help="Keep ratio of memory when the agent switch from short to long term memory. 1.0 means perfect memory, 0.0 means no memory")
    parser.add_argument('--verbose', type=bool, default=False)

    args = parser.parse_args()

    # create the environment
    maze_env = gym.make("mazeworld-v2", enable_render=False, max_steps=args.max_steps, visibility_3D=args.visibility_3D, 
                        command_in_observation=True)

    # sample the task
    task = MazeTaskSampler(verbose=True)
    maze_env.set_task(task)

    # create an smart SLAM agent
    agent = SmartSLAMAgent(maze_env=maze_env, memory_keep_ratio=args.memory_keep_ratio, render=True)

    observation, _ = maze_env.reset()
    done=False
    sum_reward = 0
    reward = 0

    while not done:
        action = agent.step(observation, reward)
        observation, reward, done, _ = maze_env.step(action)
        sum_reward += reward
        if(args.verbose):
            print("Instant r = %.2f, Accumulate r = %.2f" % (reward, sum_reward))
        if(maze_env.key_done):
            break
    print("Episode is over! You got %.2f score."%sum_reward)

    if(args.save_replay is not None):
        maze_env.save_trajectory(args.save_replay)



from numpy import random
import time

def pseudo_random_seed(hyperseed=0):
    '''
    Generate a pseudo random seed based on current time and system random number
    '''
    timestamp = time.time_ns()
    system_random = int(random.random() * 100000000)
    pseudo_random = timestamp + system_random + hyperseed
    
    return pseudo_random % (4294967296)


#!/usr/bin/env python
# coding=utf8
# File: test.py
import gym
import sys
import l3c.mazeworld
import time
from l3c.mazeworld import MazeTaskSampler, Resampler
from l3c.mazeworld.agents import SmartSLAMAgent
from numpy import random

def test_agent_maze(max_steps=1000):
    maze_env = gym.make("mazeworld-v2", enable_render=False, max_steps=max_steps)
    task = MazeTaskSampler(verbose=True)
    maze_env.set_task(Resampler(task))

    # Must intialize agent after reset
    agent = SmartSLAMAgent(maze_env=maze_env, memory_keep_ratio=0.25, render=False)

    done=False
    observation = maze_env.reset()
    sum_reward = 0
    reward = 0
    while not done:
        action = agent.step(observation, reward)
        observation, reward, done, _ = maze_env.step(action)
        loc_map = maze_env.get_local_map()
        global_map = maze_env.get_global_map()
        sum_reward += reward
    print("...Test Finishes. Get score %f, steps = %s\n\n---------\n\n"%(sum_reward, max_steps))

if __name__=="__main__":
    for _ in range(10):
        test_agent_maze(max_steps=100)
    print("\n\nCongratulations!!!\n\nAll Tests Have Been Passed\n\n")



import numpy
import math
from numba import njit

PI_4 = 0.7853981
PI_2 = 1.5707963
PI = 3.1415926
t_PI = 6.2831852
PI2d = 57.29578

OFFSET_10 = numpy.asarray([0.5, 0.5], dtype="float64")
OFFSET_01 = numpy.asarray([-0.5, 0.5], dtype="float64")
OFFSET_m0 = numpy.asarray([-0.5, -0.5], dtype="float64")
OFFSET_0m = numpy.asarray([0.5, -0.5], dtype="float64")

DEFAULT_ACTION_SPACE_16 = [(0.0, 0.5), 
                        (0.05, 0.0), (-0.05, 0.0),
                        (0.1, 0.0), (-0.1, 0.0),
                        (0.2, 0.0), (-0.2, 0.0),
                        (0.3, 0.0), (-0.3, 0.0),
                        (0.5, 0.0), (-0.5, 0.0), 
                        (0.0, 1.0),
                        (0.05, 1.0), 
                        (-0.05, 1.0), 
                        (0.10, 1.0), 
                        (-0.10, 1.0), 
                        ]

DEFAULT_ACTION_SPACE_32 = [(0.0, 0.2), 
                        (0.02, 0.0), (-0.02, 0.0),
                        (0.05, 0.0), (-0.05, 0.0),
                        (0.1, 0.0), (-0.1, 0.0),
                        (0.2, 0.0), (-0.2, 0.0),
                        (0.3, 0.0), (-0.3, 0.0),
                        (0.4, 0.0), (-0.4, 0.0),
                        (0.5, 0.0), (-0.5, 0.0), 
                        (0.0, 0.5), (0.0, 1.0),
                        (0.02, 0.5), (0.02, 1.0), 
                        (-0.02, 0.5), (-0.02, 1.0),
                        (0.05, 0.5), (0.05, 1.0), 
                        (-0.05, 0.5), (-0.05, 1.0),
                        (0.10, 0.5), (0.10, 1.0), 
                        (-0.10, 0.5), (-0.10, 1.0),
                        (0.0, -0.2),
                        (0.1, -0.2), (-0.1, -0.2)
                        ]

@njit(cache=True)
def angle_normalization(t):
    while(t > PI):
        t -= t_PI
    while(t < -PI):
        t += t_PI
    return t

@njit(cache=True)
def nearest_point(pos, line_1, line_2):
    unit_ori = line_2 - line_1
    edge_norm = numpy.sqrt(numpy.sum(unit_ori * unit_ori))
    unit_ori /= max(1.0e-6, edge_norm)

    dist_1 = numpy.sum((pos - line_1) * unit_ori)
    if(dist_1 > edge_norm):
        return numpy.sqrt(numpy.sum((pos - line_2) ** 2)), numpy.copy(line_2)
    elif(dist_1 < 0):
        return numpy.sqrt(numpy.sum((pos - line_1) ** 2)), numpy.copy(line_1)
    else:
        line_p = line_1 + dist_1 * unit_ori
        return numpy.sqrt(numpy.sum((pos - line_p) ** 2)), numpy.copy(line_p)

@njit(cache=True)
def collision_force(dist_vec, cell_size, col_dist):
    dist = float(numpy.sqrt(numpy.sum(dist_vec * dist_vec)))
    eff_col_dist = col_dist / cell_size
    if(dist > 0.708 + eff_col_dist):
        return numpy.array([0.0, 0.0], dtype="float64")
    if(abs(dist_vec[0]) < 0.5 and abs(dist_vec[1]) < 0.5):
        return numpy.float64(0.50 / max(dist, 1.0e-6) * (0.708 + eff_col_dist - dist) * cell_size) * dist_vec
    x_pos = (dist_vec[0] + dist_vec[1] > 0)
    y_pos = (dist_vec[1] - dist_vec[0] > 0)
    if(x_pos and y_pos):
        dist, np = nearest_point(dist_vec, OFFSET_10, OFFSET_01)
    elif((not x_pos) and y_pos):
        dist, np = nearest_point(dist_vec, OFFSET_01, OFFSET_m0)
    elif((not x_pos) and (not y_pos)):
        dist, np = nearest_point(dist_vec, OFFSET_m0, OFFSET_0m)
    elif(x_pos and (not y_pos)):
        dist, np = nearest_point(dist_vec, OFFSET_0m, OFFSET_10)

    if(eff_col_dist < dist):
        return numpy.array([0.0, 0.0], dtype="float64")
    else:
        ori = dist_vec - np
        ori_norm = numpy.sqrt(numpy.sum(ori * ori))
        ori *= 1.0 / max(1.0e-6, ori_norm)
        return (0.50 * (eff_col_dist - dist) * cell_size) * ori

@njit(cache=True)
def vector_move_no_collision(ori, turn_rate, walk_speed, dt):
    d_theta = turn_rate * dt
    arc = walk_speed * dt
    c_theta = numpy.cos(ori)
    s_theta = numpy.sin(ori)
    c_dt = numpy.cos(0.5 * d_theta)
    s_dt = numpy.sin(0.5 * d_theta)

    n_ori = ori + d_theta
    # Shape it to [-PI, PI]
    n_ori = angle_normalization(n_ori)

    if(abs(d_theta) < 1.0e-8):
        d_x = c_theta * arc
        d_y = s_theta * arc
    else:
        # Turning Radius
        rad = walk_speed / turn_rate
        offset = 2.0 * s_dt * rad
        c_n = c_theta * c_dt - s_theta * s_dt
        s_n = c_theta * s_dt + s_theta * c_dt
        d_x = c_n * offset
        d_y = s_n * offset

    return n_ori, numpy.array([d_x, d_y], dtype="float64") 

@njit(cache=True)
def search_optimal_action(ori, targ1, targ2, candidate_action, delta_t):
    d_targ1 = numpy.array(targ1, dtype=numpy.float64)
    if(targ2 is not None):
        d_targ2 = numpy.array(targ2, dtype=numpy.float64)
    costs = []
    for action in candidate_action:
        tr = action[0] * PI
        ws = action[1]
        n_ori, n_loc = vector_move_no_collision(ori, tr, ws, delta_t)

        # The position error costs
        dist_loss = numpy.sum((n_loc - d_targ1) ** 2)
        dist = numpy.sqrt(dist_loss)
        cost = dist_loss

        # The action costs
        cost += 1.0e-4 * (action[0] ** 2 + action[1] ** 2)

        # The orientation costs
        targ1_ang = math.atan2(d_targ1[1], d_targ1[0])
        delta1_ang = angle_normalization(targ1_ang - n_ori)
        delta2_ang = delta1_ang
        if(targ2 is not None):  # Else try prepare for the next target
            targ2_ang = math.atan2(d_targ2[1], d_targ2[0])
            delta2_ang = angle_normalization(targ2_ang - n_ori)

        # Try to face the next target by looking ahead
        f= min(dist/0.2, 1.0)
        cost += delta1_ang * delta1_ang * f + delta2_ang * delta2_ang * (1 - f)
        costs.append(cost)
    return numpy.argmin(numpy.array(costs))

def vector_move_with_collision(ori, pos, turn_rate, walk_speed, delta_t, cell_walls, cell_size, col_dist):
    slide_factor = 0.20
    tmp_pos = numpy.copy(numpy.array(pos, dtype="float64"))
    
    t_prec = 0.01
    iteration = int(delta_t / t_prec)
    collision = 0.0

    for i in range(iteration + 1):
        t_res = min(delta_t - i * t_prec, t_prec)
        if(t_res < 1.0e-8):
            continue
        ori, offset = vector_move_no_collision(ori, turn_rate, walk_speed, t_res)
        exp_pos = tmp_pos + offset
        exp_cell = exp_pos / cell_size
        
        #consider the collision in new cell
        col_f = numpy.array([0.0, 0.0], dtype="float64")
        for i in range(-1, 2): 
            for j in range(-1, 2): 
                w_i = i + int(exp_cell[0])
                w_j = j + int(exp_cell[1])
                if(w_i > -1 and w_i < cell_walls.shape[0] and w_j > -1  and w_j < cell_walls.shape[1]):
                    if(cell_walls[w_i,w_j] > 0):
                        cell_deta = exp_cell - numpy.floor(exp_cell) - numpy.array([i + 0.5, j + 0.5], dtype="float32")
                        col_f += collision_force(cell_deta, cell_size, col_dist)
        tmp_pos = col_f + exp_pos
        collision += numpy.sqrt(numpy.sum(col_f ** 2))

    return ori, tmp_pos, collision



"""
Gym Environment For Maze3D
"""
import numpy
import gym
import pygame
import time
from numpy import random
from copy import deepcopy
from l3c.utils import pseudo_random_seed
from l3c.anymdp.solver import check_task_quality, check_reward_quality, check_transition_quality

def irwin_hall(n:int, N:int):
    # Generate a random categorical distribution with N categories, which n non-zero values
    val = random.rand(n + 1)
    val[0] = 0
    val[1] = 1
    val.sort()
    val = val[1:] - val[:-1]
    random.shuffle(val)
    val = numpy.concatenate((val, numpy.zeros(N - n)))
    random.shuffle(val)
    return val

def reward_sampler(state_space:int, 
                   action_space:int, 
                   reward_sparsity:float,
                   positive_ratio:float,
                   reward_noise_max:float, 
                   reward_type='binomial'):
    reward_mask = numpy.zeros((state_space, action_space))
    while numpy.sum(reward_mask) < 1:
        reward_sparsity = min(reward_sparsity, 1.0)
        reward_sparsity = random.random() * reward_sparsity
        reward_sparsity = max(1 / (state_space*action_space), reward_sparsity) # prevent all-zeros
        reward_mask = random.binomial(1, reward_sparsity, size=(state_space, action_space))
    reward_noise = reward_noise_max * random.rand(state_space, action_space) * random.random()

    sign_mask = 2.0 * random.binomial(1, positive_ratio, size=(state_space, action_space)) - 1
    reward_matrix = numpy.abs(random.normal(loc=0, scale=1.0, size=(state_space, action_space))) * reward_mask * sign_mask
    if(reward_type == 'binomial'):
        reward_matrix = numpy.clip(reward_matrix, -1.0, 1.0)
    return reward_matrix, reward_noise

def AnyMDPTaskSampler(state_space:int=128,
                 action_space:int=5, 
                 max_reward_sparsity=0.03,
                 max_positive_reward_ratio=0.50,
                 transition_diversity = 6,
                 reward_noise_max=0.30,
                 reward_noise_type=None,
                 max_steps_average=200,
                 seed=None):
    # Sampling Transition Matrix and Reward Matrix based on Irwin-Hall Distribution and Gaussian Distribution
    if(seed is not None):
        random.seed(seed)
    else:
        random.seed(pseudo_random_seed())

    if(reward_noise_type not in ['binomial', 'normal', None]):
        raise ValueError('Reward type must be either binomial or normal')
    if(reward_noise_type is None):
        reward_noise_type = random.choice(['binomial', 'normal'])

    max_steps = max(int(numpy.random.exponential(max_steps_average)), max_steps_average // 4)
    
    n = min(max(1, int(transition_diversity * random.random())), state_space)

    # Generate Transition Matrix While Check its Quality
    transition_matrix = None
    while not check_transition_quality(transition_matrix):
        transition_matrix = numpy.zeros((state_space, action_space, state_space))
        for i in range(state_space):
            for j in range(action_space):
                fn = max(1, int(random.random() * n))  # make sparsity more random
                transition = irwin_hall(fn, state_space)
                transition /= sum(transition)
                transition_matrix[i][j] = transition
        
    # Reward Can not be too sparse
    reward_matrix = None
    while not check_reward_quality(transition_matrix, reward_matrix):
        reward_matrix = numpy.zeros((state_space, action_space))
        reward_matrix, reward_noise = reward_sampler(state_space, action_space, 
                                                 max_reward_sparsity,
                                                 random.random() * max_positive_reward_ratio,
                                                 reward_noise_max, 
                                                 reward_noise_type)
    
    return {'transition': transition_matrix, 
            'max_steps': max_steps, 
            'reward': reward_matrix, 
            'reward_noise': reward_noise, 
            'reward_noise_type': reward_noise_type}


def Resampler(task, seed=None, max_reward_sparsity=0.03, max_positive_reward_ratio=0.50, reward_noise_max=0.30):
    # Sampling Transition Matrix and Reward Matrix based on Irwin-Hall Distribution and Gaussian Distribution
    if(seed is not None):
        random.seed(seed)
    else:
        random.seed(pseudo_random_seed())

    ntask = deepcopy(task)
    state_space, action_space = task['reward'].shape
    reward_matrix = numpy.zeros((state_space, action_space))
    
    # Reward Can not be too sparse
    reward_matrix = None
    while not check_reward_quality(task["transition"], reward_matrix):
        reward_matrix, reward_noise = reward_sampler(state_space, 
                                                        action_space, 
                                                        max_reward_sparsity, 
                                                        random.random() * max_positive_reward_ratio,
                                                        reward_noise_max,
                                                        task['reward_noise_type'])
    
    ntask.update({'reward_noise': reward_noise,
                  'reward': reward_matrix})
    return ntask



"""
Core File of Maze Env
"""
import os
import numpy
import pygame
import random
import time
from pygame import font
from numpy import random as npyrnd
from numpy.linalg import norm
from l3c.mazeworld.envs.ray_caster_utils import landmarks_rgb, landmarks_color
from l3c.mazeworld.envs.dynamics import PI
from l3c.mazeworld.envs.task_sampler import MAZE_TASK_MANAGER
from l3c.mazeworld.envs.ray_caster_utils import paint_agent_arrow

class MazeBase(object):
    def __init__(self, **kw_args):
        for k in kw_args:
            self.__dict__[k] = kw_args[k]
        pygame.init()

    def set_task(self, task_config):
        # initialize textures
        self._cell_walls = numpy.copy(task_config["cell_walls"])
        self._cell_texts = task_config["cell_texts"]
        self._start = task_config["start"]
        self._n = numpy.shape(self._cell_walls)[0]
        self._cell_landmarks = task_config["cell_landmarks"]
        self._cell_size = task_config["cell_size"]
        self._wall_height = task_config["wall_height"]
        self._agent_height = task_config["agent_height"]
        self._step_reward = task_config["step_reward"]
        self._goal_reward = task_config["goal_reward"]
        self._collision_reward = task_config["collision_reward"]
        self._landmarks_coordinates = task_config["landmarks_coordinates"]
        self._commands_sequence = task_config["commands_sequence"]
        self._ground_text = task_config["ground_text"]
        self._ceiling_text = task_config["ceiling_text"]
        self._int_max = 100000000
        self._commands_maxlife = 500

        # Texture Color Used in Map
        self._text_color=[]
        for text_surf in MAZE_TASK_MANAGER.textlib_walls:
            avg_color = numpy.mean(text_surf, axis=(0,1))
            self._text_color.append(pygame.Color(int(avg_color[0]), int(avg_color[1]), int(avg_color[2])))

        assert self._agent_height < self._wall_height and self._agent_height > 0, "the agent height must be > 0 and < wall height"
        assert self._cell_walls.shape == self._cell_texts.shape, "the dimension of walls must be equal to textures"
        assert self._cell_walls.shape[0] == self._cell_walls.shape[1], "only support square shape"

    def refresh_command(self):
        """
        Update the command for selecting the target to navigate
        At the same time, update the instant_rewards
        """
        if(self._command is not None):
            x,y = self._landmarks_coordinates[self._command]
            self._instant_rewards[x, y] = 0.0

        self._commands_sequence_idx += 1
        self._commands_exists = 0
        if(self._commands_sequence_idx > len(self._commands_sequence) - 1):
            return True
        self._command = self._commands_sequence[self._commands_sequence_idx]
        x,y = self._landmarks_coordinates[self._command]
        self._instant_rewards[x,y] = self._goal_reward
        return False

    def step_limits(self):
        if(self._commands_maxlife > 0):
            if(self._commands_exists >= self._commands_maxlife):
                return True
        return False

    def reach_goal(self):
        g_x, g_y = self._landmarks_coordinates[self._command]
        goal = ((g_x == self._agent_grid[0]) and (g_y == self._agent_grid[1]))
        return goal

    def reset(self):
        self._agent_grid = numpy.copy(self._start)
        self._agent_loc = self.get_cell_center(self._start)
        self._agent_trajectory = [numpy.copy(self._agent_loc)]

        # Record all observed cells
        self._cell_exposed = numpy.zeros_like(self._cell_walls).astype(bool)

        # Maximum w and h in the space
        self._size = self._n * self._cell_size

        # Valid in 3D
        self._agent_ori = 0.0
        self._instant_rewards = numpy.zeros_like(self._cell_landmarks, dtype="float32")

        # Initialization related to tasks
        self._commands_sequence_idx = -1
        self._command = None
        self.refresh_command()

        self.update_observation()
        self.steps = 0
        return self.get_observation()

    def evaluation_rule(self):
        self.steps += 1
        self._commands_exists += 1
        self._agent_trajectory.append(numpy.copy(self._agent_loc))
        agent_grid_idx = tuple(self._agent_grid)

        reward = self._instant_rewards[agent_grid_idx] + self._step_reward
        done = False
        if(self.reach_goal() or self.step_limits()):
            done = self.refresh_command()
        done = done or self.episode_steps_limit()

        return reward, done

    def do_action(self, action):
        raise NotImplementedError()

    def render_init(self, view_size):
        """
        Initialize a God View With Landmarks
        """
        font.init()
        self._font = font.SysFont("Arial", 18)

        #Initialize the agent drawing
        self._render_cell_size = view_size / self._n
        self._view_size = view_size

        self._obs_logo = self._font.render("Observation", 0, pygame.Color("red"))

        self._screen = pygame.Surface((3 * view_size, view_size))
        self._screen = pygame.display.set_mode((3 * view_size, view_size))
        self._screen.fill(pygame.Color("white"))
        logo_gmap = self._font.render("Global Map (Keep It Hidden From Agent)", 0, pygame.Color("red"))
        logo_lmap = self._font.render("Local Map (Keep It Hidden From Agent)", 0, pygame.Color("red"))

        self._screen.blit(logo_gmap,(view_size + 90, 5))
        self._screen.blit(logo_lmap,(2 * view_size + 90, 5))

        pygame.display.set_caption("MazeWorld Render")

    def render_map(self):
        """
        Cover landmarks with white in case it is not refreshed
        """
        empty_range = 32
        gm_surf, _ = self.get_global_map(resolution=(512, 512))
        lm_surf, _ = self.get_local_map(map_range=8, resolution=(512, 512))

        gm_surf = pygame.transform.scale(gm_surf, (self._view_size - 2 * empty_range, self._view_size - 2 * empty_range))
        lm_surf = pygame.transform.scale(lm_surf, (self._view_size - 2 * empty_range, self._view_size - 2 * empty_range))

        self._screen.blit(gm_surf, (self._view_size + empty_range, empty_range))
        self._screen.blit(lm_surf, (2.0 * self._view_size + empty_range, empty_range))


    def render_observation(self):
        """
        Need to implement the logic for observation painting
        """
        raise NotImplementedError()

    def render_update(self):
        #Paint God View
        self.render_map()

        #Paint Agent and Observation
        self.render_observation()

        pygame.display.update()
        done = False
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                done=True
        keys = pygame.key.get_pressed()
        return done, keys

    def render_trajectory(self, file_name, additional=None):
        # Render god view with record on the trajectory
        if(additional is not None):
            aw, ah = (additional["surfaces"][0].get_width(),additional["surfaces"][0].get_height())
        else:
            aw, ah = (0, 0)

        traj_screen = pygame.Surface((self._view_size + aw, max(self._view_size, ah)))
        traj_screen.fill(pygame.Color("white"))
        surf_map, _ = self.get_global_map((self._view_size, self._view_size))
        traj_screen.blit(surf_map, (0, 0))

        for i in range(len(self._agent_trajectory)-1):
            factor = i / len(self._agent_trajectory)
            noise = (factor - 0.5) * 0.10
            p = numpy.array(self._agent_trajectory[i]) / self._cell_size * self._render_cell_size
            n = numpy.array(self._agent_trajectory[i + 1]) / self._cell_size * self._render_cell_size
            pygame.draw.line(traj_screen, pygame.Color(int(255 * factor), int(255 * (1 - factor)), 0, 255), 
                             (float(p[0]), float(p[1])), (float(n[0]), float(n[1])), width=2)

        # paint some additional surfaces where necessary
        if(additional != None):
            for i in range(len(additional["surfaces"])):
                traj_screen.blit(additional["surfaces"][i], (self._view_size, 0))
                pygame.image.save(traj_screen, file_name.split(".")[0] + additional["file_names"][i] + ".png")
        else:
            pygame.image.save(traj_screen, file_name)

    def episode_steps_limit(self):
        return self.steps > self.max_steps-1

    def get_cell_center(self, cell):
        p_x = cell[0] * self._cell_size + 0.5 * self._cell_size
        p_y = cell[1] * self._cell_size + 0.5 * self._cell_size
        return [p_x, p_y]

    def get_loc_grid(self, loc):
        p_x = int(loc[0] / self._cell_size)
        p_y = int(loc[1] / self._cell_size)
        return [p_x, p_y]

    def get_loc_grid_float(self, loc):
        p_x = (loc[0] / self._cell_size)
        p_y = (loc[1] / self._cell_size)
        return [p_x, p_y]

    def movement_control(self, keys):
        """
        Implement the movement control logic, or ''agent dynamics''
        """
        raise NotImplementedError()

    def update_observation(self):
        """
        Update the observation, which is used for returning the state when ''get_observation''
        """
        raise NotImplementedError()

    def get_observation(self):
        return numpy.copy(self._observation)

    def get_local_map(self, map_range=8, resolution=(128, 128)):
        if("_agent_ori" in self.__dict__):
            cos_ori = numpy.cos(self._agent_ori)
            sin_ori = numpy.sin(self._agent_ori)
        else:
            cos_ori = 1.0
            sin_ori = 0.0
        rot_mat = numpy.array([[cos_ori, sin_ori], [-sin_ori, cos_ori]])

        surf_map = pygame.Surface(resolution)
        surf_map.fill(pygame.Color("grey"))

        # Number of pixels per distance
        render_x = 0.5 * resolution[0] / map_range
        render_y = 0.5 * resolution[1] / map_range
        render_avg = 0.5 * (render_x + render_y)
        it = numpy.nditer(self._cell_walls, flags=["multi_index"])
        landmark_size = 0.60
        max_range = map_range + 0.5 * self._cell_size
        map_grids = map_range / self._cell_size

        # Number of pixels per grids
        delta = self._cell_size * numpy.array([render_x, render_y])

        for _ in it:
            x,y = it.multi_index
            p_x, p_y = self.get_cell_center([x, y])
            d_x = p_x - self._agent_loc[0]
            d_y = p_y - self._agent_loc[1]
            dx = d_x / self._cell_size
            dy = d_y / self._cell_size
            rot_p = numpy.matmul(rot_mat, numpy.array([d_x, d_y]))
            rot_g = rot_p / self._cell_size
            f_x = rot_p[0]
            f_y = rot_p[1]
            fx = rot_g[0]
            fy = rot_g[1]
            if(abs(f_x) > max_range or abs(f_y) > max_range):
                # Cell out of range, skip
                continue
            landmarks_id = self._cell_landmarks[x,y]
            p1 = (delta * (numpy.matmul(rot_mat, numpy.array([dx - 0.5, dy - 0.5])) + map_grids)).tolist()
            p2 = (delta * (numpy.matmul(rot_mat, numpy.array([dx - 0.5, dy + 0.5])) + map_grids)).tolist()
            p3 = (delta * (numpy.matmul(rot_mat, numpy.array([dx + 0.5, dy + 0.5])) + map_grids)).tolist()
            p4 = (delta * (numpy.matmul(rot_mat, numpy.array([dx + 0.5, dy - 0.5])) + map_grids)).tolist()
            if(self._cell_walls[x,y] > 0):
                pygame.draw.polygon(surf_map, self._text_color[self._cell_texts[x,y]],
                                 [p1, p2, p3, p4])
            else:
                pygame.draw.polygon(surf_map, pygame.Color("white"),
                                 [p1, p2, p3, p4])
            if(landmarks_id > -1):
                pygame.draw.circle(surf_map, landmarks_color(landmarks_id, opacity=0.0), 
                                 ((fx + map_grids) * delta[0], (fy + map_grids) * delta[1]), 0.5 * landmark_size * render_avg * self._cell_size, 
                                 width=0)
        npy_map = pygame.surfarray.array3d(surf_map)
        return surf_map, npy_map

    def get_global_map(self, resolution=(128, 128)):
        surf_map = pygame.Surface(resolution)
        surf_map.fill(pygame.Color("white"))
        render_x = resolution[0] / self._n
        render_y = resolution[1] / self._n
        render_avg = 0.5 * (render_x + render_y)
        it = numpy.nditer(self._cell_walls, flags=["multi_index"])
        landmark_size = 0.60
        landmark_emp = 0.5 * (1.0 - landmark_size)

        for _ in it:
            x,y = it.multi_index
            landmarks_id = self._cell_landmarks[x,y]
            if(self._cell_walls[x,y] > 0):
                pygame.draw.rect(surf_map, self._text_color[self._cell_texts[x,y]],
                                 (x * render_x, y * render_y, render_x, render_y))
            if(landmarks_id > -1):
                pygame.draw.circle(surf_map, landmarks_color(landmarks_id, opacity=0.0), 
                                 ((x + 0.5) * render_x, (y + 0.5) * render_y), 0.5 * landmark_size * render_avg, 
                                 width=0)
        pos_conversion = numpy.array([render_x, render_y]) / self._cell_size
        agent_pos = numpy.array(self._agent_loc) * pos_conversion
        paint_agent_arrow(surf_map, pygame.Color("gray"), (0, 0), (agent_pos[0], agent_pos[1]), self._agent_ori, 
                0.4 * render_avg, 0.5 * render_avg)
        npy_map = pygame.surfarray.array3d(surf_map)
        return surf_map, npy_map
